@@ -5,27 +5,27 @@
  * libradius.h	Structures and prototypes
  *		for the radius library.
  *
- * Version:	$Id: libradius.h,v 1.154 2008/05/10 05:37:22 aland Exp $
+ * Version:	$Id: b598f74ef948b65b60d116d2ed5c947372ef4b0e $
  *
- *   This program is free software; you can redistribute it and/or modify
- *   it under the terms of the GNU General Public License as published by
- *   the Free Software Foundation; either version 2 of the License, or
- *   (at your option) any later version.
+ *   This library is free software; you can redistribute it and/or
+ *   modify it under the terms of the GNU Lesser General Public
+ *   License as published by the Free Software Foundation; either
+ *   version 2.1 of the License, or (at your option) any later version.
  *
- *   This program is distributed in the hope that it will be useful,
+ *   This library is distributed in the hope that it will be useful,
  *   but WITHOUT ANY WARRANTY; without even the implied warranty of
- *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *   GNU General Public License for more details.
+ *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
+ *   Lesser General Public License for more details.
  *
- *   You should have received a copy of the GNU General Public License
- *   along with this program; if not, write to the Free Software
+ *   You should have received a copy of the GNU Lesser General Public
+ *   License along with this library; if not, write to the Free Software
  *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
  *
  * Copyright 1999,2000,2001,2002,2003,2004,2005,2006,2007,2008  The FreeRADIUS server project
  */
 
 #include <freeradius-devel/ident.h>
-RCSIDH(libradius_h, "$Id: libradius.h,v 1.154 2008/05/10 05:37:22 aland Exp $")
+RCSIDH(libradius_h, "$Id: b598f74ef948b65b60d116d2ed5c947372ef4b0e $")
 
 #include <freeradius-devel/missing.h>
 
@@ -36,7 +36,9 @@ RCSIDH(libradius_h, "$Id: libradius.h,v 1.154 2008/05/10 05:37:22 aland Exp $")
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
+#include <signal.h>
 
+#include <freeradius-devel/threads.h>
 #include <freeradius-devel/radius.h>
 #include <freeradius-devel/token.h>
 #include <freeradius-devel/hash.h>
@@ -47,30 +49,48 @@ RCSIDH(libradius_h, "$Id: libradius.h,v 1.154 2008/05/10 05:37:22 aland Exp $")
 #endif
 #endif
 
+#ifndef FALSE
+#define FALSE 0
+#endif
+#ifndef TRUE
+/*
+ *	This definition of true as NOT false is definitive. :) Making
+ *	it '1' can cause problems on stupid platforms.  See articles
+ *	on C portability for more information.
+ */
+#define TRUE (!FALSE)
+#endif
+
 /*
  *  Include for modules.
  */
 #include <freeradius-devel/sha1.h>
 #include <freeradius-devel/md4.h>
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
+#ifndef HAVE_SIG_T
+typedef void (*sig_t)(int);
+#endif
+
 #define EAP_START               2
 
 #define AUTH_VECTOR_LEN		16
 #define CHAP_VALUE_LENGTH       16
 #define MAX_STRING_LEN		254	/* RFC2138: string 0-253 octets */
 
-#  define VENDOR(x)		((x >> 16) & 0x7fff)
+#  define VENDOR(x)		((x >> 16) & 0xffff)
 
 #ifdef _LIBRADIUS
 #  define AUTH_HDR_LEN		20
 #  define VENDORPEC_USR		429
 #define VENDORPEC_LUCENT	4846
 #define VENDORPEC_STARENT	8164
-#  define DEBUG			if (librad_debug && fr_log_fp) fr_printf_log
-#  define debug_pair(vp)	do { if (librad_debug && fr_log_fp) { \
-					fputc('\t', fr_log_fp); \
+#  define DEBUG			if (fr_debug_flag && fr_log_fp) fr_printf_log
+#  define debug_pair(vp)	do { if (fr_debug_flag && fr_log_fp) { \
 					vp_print(fr_log_fp, vp); \
-					fputc('\n', fr_log_fp); \
 				     } \
 				} while(0)
 #  define TAG_VALID(x)          ((x) > 0 && (x) < 0x20)
@@ -98,6 +118,9 @@ typedef struct attr_flags {
 	unsigned int		array : 1; /* pack multiples into 1 attr */
 	unsigned int		has_value : 1; /* has a value */
 	unsigned int		has_value_alias : 1; /* has a value alias */
+	unsigned int		has_tlv : 1; /* has sub attributes */
+	unsigned int		is_tlv : 1; /* is a sub attribute */
+	unsigned int		encoded : 1; /* has been put into packet */
 
 	int8_t			tag;	      /* tag for tunneled attributes */
 	uint8_t		        encrypt;      /* encryption method */
@@ -111,16 +134,18 @@ typedef struct attr_flags {
 #define FLAG_ENCRYPT_TUNNEL_PASSWORD (2)
 #define FLAG_ENCRYPT_ASCEND_SECRET   (3)
 
+extern const FR_NAME_NUMBER dict_attr_types[];
+
 typedef struct dict_attr {
-	char			name[40];
-	int			attr;
+	unsigned int		attr;
 	int			type;
 	int			vendor;
         ATTR_FLAGS              flags;
+	char			name[1];
 } DICT_ATTR;
 
 typedef struct dict_value {
-	int			attr;
+	unsigned int		attr;
 	int			value;
 	char			name[1];
 } DICT_VALUE;
@@ -129,6 +154,7 @@ typedef struct dict_vendor {
 	int			vendorpec;
 	int			type; /* length of type data */
 	int			length;	/* length of length data */
+	int			flags;
 	char			name[1];
 } DICT_VENDOR;
 
@@ -139,10 +165,12 @@ typedef union value_pair_data {
 	struct in6_addr		ipv6addr;
 	uint32_t		date;
 	uint32_t		integer;
+	int32_t			sinteger;
 	uint8_t			filter[32];
 	uint8_t			ifid[8]; /* struct? */
 	uint8_t			ipv6prefix[18]; /* struct? */
      	uint8_t			ether[6];
+	uint8_t			*tlv;
 } VALUE_PAIR_DATA;
 
 typedef struct value_pair {
@@ -151,7 +179,16 @@ typedef struct value_pair {
 	int			vendor;
 	int			type;
 	size_t			length; /* of data */
+#ifdef __cplusplus
+	/*
+	 *	C++ hackery.  The server and modules are all C, so
+	 *	the defs here don't affect them.  But any C++ code
+	 *	gets excited over "operator", so we change the name.
+	 */
+	FR_TOKEN		op_token;
+#else
 	FR_TOKEN		operator;
+#endif
         ATTR_FLAGS              flags;
 	struct value_pair	*next;
 	uint32_t		lvalue;
@@ -164,6 +201,8 @@ typedef struct value_pair {
 #define vp_ipv6prefix data.ipv6prefix
 #define vp_filter     data.filter
 #define vp_ether      data.ether
+#define vp_signed     data.sinteger
+#define vp_tlv	      data.tlv
 
 #if 0
 #define vp_ipaddr     data.ipaddr.s_addr
@@ -186,6 +225,7 @@ typedef struct fr_ipaddr_t {
 		struct in_addr	ip4addr;
 		struct in6_addr ip6addr; /* maybe defined in missing.h */
 	} ipaddr;
+	uint32_t	scope;	/* for IPv6 */
 } fr_ipaddr_t;
 
 /*
@@ -217,7 +257,8 @@ typedef struct radius_packet {
 /*
  *	Printing functions.
  */
-void		librad_safeprint(char *in, size_t inlen,
+int		fr_utf8_char(const uint8_t *str);
+size_t		fr_print_string(const char *in, size_t inlen,
 				 char *out, size_t outlen);
 int     	vp_prints_value(char *out, size_t outlen,
 				VALUE_PAIR *vp, int delimitst);
@@ -235,10 +276,11 @@ int		dict_addattr(const char *name, int vendor, int type, int value, ATTR_FLAGS
 int		dict_addvalue(const char *namestr, const char *attrstr, int value);
 int		dict_init(const char *dir, const char *fn);
 void		dict_free(void);
-DICT_ATTR	*dict_attrbyvalue(int attr);
+DICT_ATTR	*dict_attrbyvalue(unsigned int attr);
 DICT_ATTR	*dict_attrbyname(const char *attr);
-DICT_VALUE	*dict_valbyattr(int attr, int val);
-DICT_VALUE	*dict_valbyname(int attr, const char *val);
+DICT_VALUE	*dict_valbyattr(unsigned int attr, int val);
+DICT_VALUE	*dict_valbyname(unsigned int attr, const char *val);
+const char	*dict_valnamebyattr(unsigned int attr, int value);
 int		dict_vendorbyname(const char *name);
 DICT_VENDOR	*dict_vendorbyvalue(int vendor);
 
@@ -250,14 +292,12 @@ DICT_VENDOR	*dict_vendorbyvalue(int vendor);
 #endif
 
 /* get around diffrent ctime_r styles */
-#ifdef CTIMERSTYLE
-#if CTIMERSTYLE == SOLARISSTYLE
+#if defined(CTIMERSTYLE) && ( CTIMERSTYLE == SOLARISSTYLE)
 #define CTIME_R(a,b,c) ctime_r(a,b,c)
+#define ASCTIME_R(a,b,c) asctime_r(a,b,c)
 #else
 #define CTIME_R(a,b,c) ctime_r(a,b)
-#endif
-#else
-#define CTIME_R(a,b,c) ctime_r(a,b)
+#define ASCTIME_R(a,b,c) asctime_r(a,b)
 #endif
 
 /* md5.c */
@@ -291,7 +331,9 @@ int		rad_encode(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 int		rad_sign(RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 			 const char *secret);
 
+int rad_digest_cmp(const uint8_t *a, const uint8_t *b, size_t length);
 RADIUS_PACKET	*rad_alloc(int newvector);
+RADIUS_PACKET	*rad_alloc_reply(RADIUS_PACKET *);
 void		rad_free(RADIUS_PACKET **);
 int		rad_pwencode(char *encpw, size_t *len, const char *secret,
 			     const uint8_t *vector);
@@ -306,12 +348,14 @@ int		rad_chap_encode(RADIUS_PACKET *packet, uint8_t *output,
 VALUE_PAIR	*rad_attr2vp(const RADIUS_PACKET *packet, const RADIUS_PACKET *original,
 			     const char *secret, int attribute, int length,
 			     const uint8_t *data);
+ssize_t		rad_vp2data(const VALUE_PAIR *vp, uint8_t *out, size_t outlen);
 int		rad_vp2attr(const RADIUS_PACKET *packet,
 			    const RADIUS_PACKET *original, const char *secret,
 			    const VALUE_PAIR *vp, uint8_t *ptr);
 
 /* valuepair.c */
 VALUE_PAIR	*pairalloc(DICT_ATTR *da);
+VALUE_PAIR	*paircreate_raw(int attr, int type, VALUE_PAIR *vp);
 VALUE_PAIR	*paircreate(int attr, int type);
 void		pairfree(VALUE_PAIR **);
 void            pairbasicfree(VALUE_PAIR *pair);
@@ -335,22 +379,25 @@ VALUE_PAIR     *readvp2(FILE *fp, int *pfiledone, const char *errprefix);
  *	Error functions.
  */
 #ifdef _LIBRADIUS
-void		librad_log(const char *, ...)
+void		fr_strerror_printf(const char *, ...)
 #ifdef __GNUC__
 		__attribute__ ((format (printf, 1, 2)))
 #endif
 ;
 #endif
-void		librad_perror(const char *, ...)
+void		fr_perror(const char *, ...)
 #ifdef __GNUC__
 		__attribute__ ((format (printf, 1, 2)))
 #endif
 ;
-extern char	librad_errstr[];
-extern int	librad_dodns;	/* 0 = no dns lookups */
-extern int	librad_debug;	/* 0 = no debugging information */
-extern int	librad_max_attributes; /* per incoming packet */
+extern const char *fr_strerror(void);
+extern int	fr_dns_lookups;	/* 0 = no dns lookups */
+extern int	fr_debug_flag;	/* 0 = no debugging information */
+extern int	fr_max_attributes; /* per incoming packet */
+#define	FR_MAX_PACKET_CODE (52)
+extern const char *fr_packet_codes[FR_MAX_PACKET_CODE];
 extern FILE	*fr_log_fp;
+extern void rad_print_hex(RADIUS_PACKET *packet);
 void		fr_printf_log(const char *, ...)
 #ifdef __GNUC__
 		__attribute__ ((format (printf, 1, 2)))
@@ -360,6 +407,8 @@ void		fr_printf_log(const char *, ...)
 /*
  *	Several handy miscellaneous functions.
  */
+int		fr_fault_setup(char const *cmd, char const *program);
+int		fr_set_signal(int sig, sig_t func);
 const char *	ip_ntoa(char *, uint32_t);
 char		*ifid_ntoa(char *buffer, size_t size, uint8_t *ifid);
 uint8_t		*ifid_aton(const char *ifid_str, uint8_t *ifid);
@@ -381,13 +430,16 @@ int fr_ipaddr_cmp(const fr_ipaddr_t *a, const fr_ipaddr_t *b);
 
 int		ip_hton(const char *src, int af, fr_ipaddr_t *dst);
 const char	*ip_ntoh(const fr_ipaddr_t *src, char *dst, size_t cnt);
-
+int fr_ipaddr2sockaddr(const fr_ipaddr_t *ipaddr, int port,
+		       struct sockaddr_storage *sa, socklen_t *salen);
+int fr_sockaddr2ipaddr(const struct sockaddr_storage *sa, socklen_t salen,
+		       fr_ipaddr_t *ipaddr, int * port);
 
 
 #ifdef ASCEND_BINARY
 /* filters.c */
 int		ascend_parse_filter(VALUE_PAIR *pair);
-void		print_abinary(VALUE_PAIR *vp, char *buffer, size_t len);
+void		print_abinary(const VALUE_PAIR *vp, char *buffer, size_t len, int delimitst);
 #endif /*ASCEND_BINARY*/
 
 /* random numbers in isaac.c */
@@ -456,6 +508,10 @@ void *fr_fifo_pop(fr_fifo_t *fi);
 void *fr_fifo_peek(fr_fifo_t *fi);
 int fr_fifo_num_elements(fr_fifo_t *fi);
 
+#ifdef __cplusplus
+}
+#endif
+
 #include <freeradius-devel/packet.h>
 
 #endif /*LIBRADIUS_H*/
