@@ -1,7 +1,7 @@
 /*
  * threads.c	request threading support
  *
- * Version:	$Id$
+ * Version:	$Id: f1514d177e207168e794ede0fdaeda2763497206 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: f1514d177e207168e794ede0fdaeda2763497206 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/rad_assert.h>
@@ -79,7 +79,7 @@ RCSID("$Id$")
  *  the current thread.
  *
  *  pthread_id     pthread id
- *  thread_num     server thread number, 1...number of threads
+ *  thread_id     server thread number, 1...number of threads
  *  semaphore     used to block the thread until a request comes in
  *  status        is the thread running or exited?
  *  request_count the number of requests that this thread has handled
@@ -89,7 +89,7 @@ typedef struct THREAD_HANDLE {
 	struct THREAD_HANDLE *prev;
 	struct THREAD_HANDLE *next;
 	pthread_t            pthread_id;
-	int                  thread_num;
+	int                  thread_id;
 	int                  status;
 	unsigned int         request_count;
 	time_t               timestamp;
@@ -120,9 +120,10 @@ typedef struct THREAD_POOL {
 	THREAD_HANDLE *head;
 	THREAD_HANDLE *tail;
 
-	int total_threads;
 	int active_threads;	/* protected by queue_mutex */
-	int max_thread_num;
+	int exited_threads;
+	int total_threads;
+	int max_thread_id;
 	int start_threads;
 	int max_threads;
 	int min_spare_threads;
@@ -131,6 +132,7 @@ typedef struct THREAD_POOL {
 	unsigned long request_count;
 	time_t time_last_spawned;
 	int cleanup_delay;
+	int stop_flag;
 	int spawn_flag;
 
 #ifdef WNOHANG
@@ -157,6 +159,7 @@ typedef struct THREAD_POOL {
 static THREAD_POOL thread_pool;
 static int pool_initialized = FALSE;
 static time_t last_cleaned = 0;
+static time_t almost_now = 0;
 
 static void thread_pool_manage(time_t now);
 
@@ -263,12 +266,13 @@ static void reap_children(void)
 	pthread_mutex_lock(&thread_pool.wait_mutex);
 
 	do {
+	retry:
 		pid = waitpid(0, &status, WNOHANG);
 		if (pid <= 0) break;
 
 		mytf.pid = pid;
 		tf = fr_hash_table_finddata(thread_pool.waiters, &mytf);
-		if (!tf) continue;
+		if (!tf) goto retry;
 
 		tf->status = status;
 		tf->exited = 1;
@@ -290,20 +294,46 @@ static int request_enqueue(REQUEST *request, RAD_REQUEST_FUNP fun)
 {
 	request_queue_t *entry;
 
+	/*
+	 *	If we haven't checked the number of child threads
+	 *	in a while, OR if the thread pool appears to be full,
+	 *	go manage it.
+	 */
+	if ((last_cleaned < request->timestamp) ||
+	    (thread_pool.active_threads == thread_pool.total_threads) ||
+	    (thread_pool.exited_threads > 0)) {
+		thread_pool_manage(request->timestamp);
+	}
+
 	pthread_mutex_lock(&thread_pool.queue_mutex);
 
 	thread_pool.request_count++;
 
 	if (thread_pool.num_queued >= thread_pool.max_queue_size) {
+		int complain = FALSE;
+		time_t now;
+		static time_t last_complained = 0;
+
+		now = time(NULL);
+		if (last_complained != now) {
+			last_complained = now;
+			complain = TRUE;
+		}
+		    
 		pthread_mutex_unlock(&thread_pool.queue_mutex);
 
 		/*
 		 *	Mark the request as done.
 		 */
-		radlog(L_ERR, "!!! ERROR !!! The server is blocked: discarding new request %d", request->number);
+		if (complain) {
+			radlog(L_ERR, "Something is blocking the server.  There are %d packets in the queue, waiting to be processed.  Ignoring the new request.", thread_pool.max_queue_size);
+		}
 		request->child_state = REQUEST_DONE;
 		return 0;
 	}
+	request->child_state = REQUEST_QUEUED;
+	request->component = "<core>";
+	request->module = "<queue>";
 
 	entry = rad_malloc(sizeof(*entry));
 	entry->request = request;
@@ -343,6 +373,7 @@ static int request_enqueue(REQUEST *request, RAD_REQUEST_FUNP fun)
  */
 static int request_dequeue(REQUEST **request, RAD_REQUEST_FUNP *fun)
 {
+	int blocked;
 	RAD_LISTEN_TYPE i, start;
 	request_queue_t *entry;
 
@@ -406,6 +437,9 @@ static int request_dequeue(REQUEST **request, RAD_REQUEST_FUNP *fun)
 	rad_assert((*request)->magic == REQUEST_MAGIC);
 	rad_assert(*fun != NULL);
 
+	(*request)->component = "<core>";
+	(*request)->module = "<thread>";
+
 	/*
 	 *	If the request has sat in the queue for too long,
 	 *	kill it.
@@ -415,10 +449,29 @@ static int request_dequeue(REQUEST **request, RAD_REQUEST_FUNP *fun)
 	 *	have acknowledged it as "done".
 	 */
 	if ((*request)->master_state == REQUEST_STOP_PROCESSING) {
+		(*request)->module = "<done>";
 		(*request)->child_state = REQUEST_DONE;
 		goto retry;
 	}
 
+	/*
+	 *	Produce messages for people who have 10 million rows
+	 *	in a database, without indexes.
+	 */
+	rad_assert(almost_now != 0);
+	blocked = almost_now - (*request)->timestamp;
+	if (blocked < 5) {
+		blocked = 0;
+	} else {
+		static time_t last_complained = 0;
+		
+		if (last_complained != almost_now) {
+			last_complained = almost_now;
+		} else {
+			blocked = 0;
+		}
+	}
+
 	/*
 	 *	The thread is currently processing a request.
 	 */
@@ -426,10 +479,14 @@ static int request_dequeue(REQUEST **request, RAD_REQUEST_FUNP *fun)
 
 	pthread_mutex_unlock(&thread_pool.queue_mutex);
 
+	if (blocked) {
+		radlog(L_ERR, "Request %u has been waiting in the processing queue for %d seconds.  Check that all databases are running properly!",
+		       (*request)->number, blocked);
+	}
+
 	return 1;
 }
 
-
 /*
  *	The main thread handler for requests.
  *
@@ -448,7 +505,7 @@ static void *request_handler_thread(void *arg)
 		 *	Wait to be signalled.
 		 */
 		DEBUG2("Thread %d waiting to be assigned a request",
-		       self->thread_num);
+		       self->thread_id);
 	re_wait:
 		if (sem_wait(&thread_pool.semaphore) != 0) {
 			/*
@@ -457,15 +514,15 @@ static void *request_handler_thread(void *arg)
 			 *	text.
 			 */
 			if (errno == EINTR) {
-				DEBUG2("Re-wait %d", self->thread_num);
+				DEBUG2("Re-wait %d", self->thread_id);
 				goto re_wait;
 			}
 			radlog(L_ERR, "Thread %d failed waiting for semaphore: %s: Exiting\n",
-			       self->thread_num, strerror(errno));
+			       self->thread_id, strerror(errno));
 			break;
 		}
 
-		DEBUG2("Thread %d got semaphore", self->thread_num);
+		DEBUG2("Thread %d got semaphore", self->thread_id);
 
 #ifdef HAVE_OPENSSL_ERR_H
  		/*
@@ -474,6 +531,12 @@ static void *request_handler_thread(void *arg)
 		ERR_clear_error ();
 #endif
 
+		/*
+		 *	The server is exiting.  Don't dequeue any
+		 *	requests.
+		 */
+		if (thread_pool.stop_flag) break;
+
 		/*
 		 *	Try to grab a request from the queue.
 		 *
@@ -482,14 +545,16 @@ static void *request_handler_thread(void *arg)
 		 */
 		if (!request_dequeue(&self->request, &fun)) continue;
 
-		self->request->child_pid = self->pthread_id;
+		self->request->thread_id = self->thread_id;
 		self->request_count++;
 
 		DEBUG2("Thread %d handling request %d, (%d handled so far)",
-		       self->thread_num, self->request->number,
+		       self->thread_id, self->request->number,
 		       self->request_count);
 
+		self->request->module = "";
 		radius_handle_request(self->request, fun);
+		self->request = NULL;
 
 		/*
 		 *	Update the active threads.
@@ -498,9 +563,20 @@ static void *request_handler_thread(void *arg)
 		rad_assert(thread_pool.active_threads > 0);
 		thread_pool.active_threads--;
 		pthread_mutex_unlock(&thread_pool.queue_mutex);
+
+		/*
+		 *	If the thread has handled too many requests, then make it
+		 *	exit.
+		 */
+		if ((thread_pool.max_requests_per_thread > 0) &&
+		    (self->request_count >= thread_pool.max_requests_per_thread)) {
+			DEBUG2("Thread %d handled too many requests",
+			       self->thread_id);
+			break;
+		}
 	} while (self->status != THREAD_CANCELLED);
 
-	DEBUG2("Thread %d exiting...", self->thread_num);
+	DEBUG2("Thread %d exiting...", self->thread_id);
 
 #ifdef HAVE_OPENSSL_ERR_H
 	/*
@@ -511,6 +587,10 @@ static void *request_handler_thread(void *arg)
 	ERR_remove_state(0);
 #endif
 
+	pthread_mutex_lock(&thread_pool.queue_mutex);
+	thread_pool.exited_threads++;
+	pthread_mutex_unlock(&thread_pool.queue_mutex);
+
 	/*
 	 *  Do this as the LAST thing before exiting.
 	 */
@@ -534,7 +614,7 @@ static void delete_thread(THREAD_HANDLE *handle)
 
 	rad_assert(handle->request == NULL);
 
-	DEBUG2("Deleting thread %d", handle->thread_num);
+	DEBUG2("Deleting thread %d", handle->thread_id);
 
 	prev = handle->prev;
 	next = handle->next;
@@ -575,7 +655,6 @@ static THREAD_HANDLE *spawn_thread(time_t now)
 {
 	int rcode;
 	THREAD_HANDLE *handle;
-	pthread_attr_t attr;
 
 	/*
 	 *	Ensure that we don't spawn too many threads.
@@ -592,43 +671,32 @@ static THREAD_HANDLE *spawn_thread(time_t now)
 	memset(handle, 0, sizeof(THREAD_HANDLE));
 	handle->prev = NULL;
 	handle->next = NULL;
-	handle->thread_num = thread_pool.max_thread_num++;
+	handle->thread_id = thread_pool.max_thread_id++;
 	handle->request_count = 0;
 	handle->status = THREAD_RUNNING;
 	handle->timestamp = time(NULL);
 
 	/*
-	 *	Initialize the thread's attributes to detached.
-	 *
-	 *	We could call pthread_detach() later, but if the thread
-	 *	exits between the create & detach calls, it will need to
-	 *	be joined, which will never happen.
-	 */
-	pthread_attr_init(&attr);
-	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
-
-	/*
-	 *	Create the thread detached, so that it cleans up it's
-	 *	own memory when it exits.
+	 *	Create the thread joinable, so that it can be cleaned up
+	 *	using pthread_join().
 	 *
 	 *	Note that the function returns non-zero on error, NOT
 	 *	-1.  The return code is the error, and errno isn't set.
 	 */
-	rcode = pthread_create(&handle->pthread_id, &attr,
+	rcode = pthread_create(&handle->pthread_id, 0,
 			request_handler_thread, handle);
 	if (rcode != 0) {
 		radlog(L_ERR, "Thread create failed: %s",
 		       strerror(rcode));
 		return NULL;
 	}
-	pthread_attr_destroy(&attr);
 
 	/*
 	 *	One more thread to go into the list.
 	 */
 	thread_pool.total_threads++;
 	DEBUG2("Thread spawned new child %d. Total threads in pool: %d",
-			handle->thread_num, thread_pool.total_threads);
+			handle->thread_id, thread_pool.total_threads);
 
 	/*
 	 *	Add the thread handle to the tail of the thread pool list.
@@ -693,7 +761,7 @@ static int pid_cmp(const void *one, const void *two)
  *
  *	FIXME: What to do on a SIGHUP???
  */
-int thread_pool_init(CONF_SECTION *cs, int spawn_flag)
+int thread_pool_init(CONF_SECTION *cs, int *spawn_flag)
 {
 	int		i, rcode;
 	CONF_SECTION	*pool_cf;
@@ -701,53 +769,50 @@ int thread_pool_init(CONF_SECTION *cs, int spawn_flag)
 
 	now = time(NULL);
 
-	/*
-	 *	We're not spawning new threads, don't do
-	 *	anything.
-	 */
-	if (!spawn_flag) return 0;
-
-	/*
-	 *	After a SIGHUP, we don't over-write the previous values.
-	 */
-	if (!pool_initialized) {
-		/*
-		 *	Initialize the thread pool to some reasonable values.
-		 */
-		memset(&thread_pool, 0, sizeof(THREAD_POOL));
-		thread_pool.head = NULL;
-		thread_pool.tail = NULL;
-		thread_pool.total_threads = 0;
-		thread_pool.max_thread_num = 1;
-		thread_pool.cleanup_delay = 5;
-		thread_pool.spawn_flag = spawn_flag;
-
-#ifdef WNOHANG
-		if ((pthread_mutex_init(&thread_pool.wait_mutex,NULL) != 0)) {
-			radlog(L_ERR, "FATAL: Failed to initialize wait mutex: %s",
-			       strerror(errno));
-			return -1;
-		}
-
-		/*
-		 *	Create the hash table of child PID's
-		 */
-		thread_pool.waiters = fr_hash_table_create(pid_hash,
-							     pid_cmp,
-							     free);
-		if (!thread_pool.waiters) {
-			radlog(L_ERR, "FATAL: Failed to set up wait hash");
-			return -1;
-		}
-#endif
-	}
+	rad_assert(spawn_flag != NULL);
+	rad_assert(*spawn_flag == TRUE);
+	rad_assert(pool_initialized == FALSE); /* not called on HUP */
 
 	pool_cf = cf_subsection_find_next(cs, NULL, "thread");
-	if (!pool_cf) {
-		radlog(L_ERR, "FATAL: Attempting to start in multi-threaded mode with no thread configuration in radiusd.conf");
+	if (!pool_cf) *spawn_flag = FALSE;
+
+	/*
+	 *	Initialize the thread pool to some reasonable values.
+	 */
+	memset(&thread_pool, 0, sizeof(THREAD_POOL));
+	thread_pool.head = NULL;
+	thread_pool.tail = NULL;
+	thread_pool.total_threads = 0;
+	thread_pool.max_thread_id = 1;
+	thread_pool.cleanup_delay = 5;
+	thread_pool.stop_flag = 0;
+	thread_pool.spawn_flag = *spawn_flag;
+	
+	/*
+	 *	Don't bother initializing the mutexes or
+	 *	creating the hash tables.  They won't be used.
+	 */
+	if (!*spawn_flag) return 0;
+	
+#ifdef WNOHANG
+	if ((pthread_mutex_init(&thread_pool.wait_mutex,NULL) != 0)) {
+		radlog(L_ERR, "FATAL: Failed to initialize wait mutex: %s",
+		       strerror(errno));
 		return -1;
 	}
 
+	/*
+	 *	Create the hash table of child PID's
+	 */
+	thread_pool.waiters = fr_hash_table_create(pid_hash,
+						   pid_cmp,
+						   free);
+	if (!thread_pool.waiters) {
+		radlog(L_ERR, "FATAL: Failed to set up wait hash");
+		return -1;
+	}
+#endif
+
 	if (cf_section_parse(pool_cf, NULL, thread_config) < 0) {
 		return -1;
 	}
@@ -761,6 +826,16 @@ int thread_pool_init(CONF_SECTION *cs, int spawn_flag)
 		thread_pool.max_spare_threads = 1;
 	if (thread_pool.max_spare_threads < thread_pool.min_spare_threads)
 		thread_pool.max_spare_threads = thread_pool.min_spare_threads;
+	if ((thread_pool.max_queue_size < 2) || (thread_pool.max_queue_size > 1048576)) {
+		radlog(L_ERR, "FATAL: max_queue_size value must be in range 2-1048576");
+		return -1;
+	}
+
+	if (thread_pool.start_threads > thread_pool.max_threads) {
+		radlog(L_ERR, "FATAL: start_servers (%i) must be <= max_servers (%i)",
+		       thread_pool.start_threads, thread_pool.max_threads);
+		return -1;
+	}
 
 	/*
 	 *	The pool has already been initialized.  Don't spawn
@@ -792,7 +867,7 @@ int thread_pool_init(CONF_SECTION *cs, int spawn_flag)
 	 *	Allocate multiple fifos.
 	 */
 	for (i = 0; i < RAD_LISTEN_MAX; i++) {
-		thread_pool.fifo[i] = fr_fifo_create(65536, NULL);
+		thread_pool.fifo[i] = fr_fifo_create(thread_pool.max_queue_size, NULL);
 		if (!thread_pool.fifo[i]) {
 			radlog(L_ERR, "FATAL: Failed to set up request fifo");
 			return -1;
@@ -828,6 +903,40 @@ int thread_pool_init(CONF_SECTION *cs, int spawn_flag)
 }
 
 
+/*
+ *	Stop all threads in the pool.
+ */
+void thread_pool_stop(void)
+{
+	int i;
+	int total_threads;
+	THREAD_HANDLE *handle;
+	THREAD_HANDLE *next;
+
+	/*
+	 *	Set pool stop flag.
+	 */
+	thread_pool.stop_flag = 1;
+
+	/*
+	 *	Wakeup all threads to make them see stop flag.
+	 */
+	total_threads = thread_pool.total_threads;
+	for (i = 0; i != total_threads; i++) {
+		sem_post(&thread_pool.semaphore);
+	}
+
+	/*
+	 *	Join and free all threads.
+	 */
+	for (handle = thread_pool.head; handle; handle = next) {
+		next = handle->next;
+		pthread_join(handle->pthread_id, NULL);
+		delete_thread(handle);
+	}
+}
+
+
 /*
  *	Assign a new request to a free thread.
  *
@@ -836,7 +945,7 @@ int thread_pool_init(CONF_SECTION *cs, int spawn_flag)
  */
 int thread_pool_addrequest(REQUEST *request, RAD_REQUEST_FUNP fun)
 {
-	time_t now = request->timestamp;
+	almost_now = request->timestamp;
 
 	/*
 	 *	We've been told not to spawn threads, so don't.
@@ -865,9 +974,9 @@ int thread_pool_addrequest(REQUEST *request, RAD_REQUEST_FUNP fun)
 	 *	in a while, OR if the thread pool appears to be full,
 	 *	go manage it.
 	 */
-	if ((last_cleaned < now) ||
+	if ((last_cleaned < almost_now) ||
 	    (thread_pool.active_threads == thread_pool.total_threads)) {
-		thread_pool_manage(now);
+		thread_pool_manage(almost_now);
 	}
 
 	return 1;
@@ -886,6 +995,25 @@ static void thread_pool_manage(time_t now)
 	THREAD_HANDLE *handle, *next;
 	int active_threads;
 
+	/*
+	 *	Loop over the thread pool, deleting exited threads.
+	 */
+	for (handle = thread_pool.head; handle; handle = next) {
+		next = handle->next;
+
+		/*
+		 *	Maybe we've asked the thread to exit, and it
+		 *	has agreed.
+		 */
+		if (handle->status == THREAD_EXITED) {
+			pthread_join(handle->pthread_id, NULL);
+			delete_thread(handle);
+			pthread_mutex_lock(&thread_pool.queue_mutex);
+			thread_pool.exited_threads--;
+			pthread_mutex_unlock(&thread_pool.queue_mutex);
+		}
+	}
+
 	/*
 	 *	We don't need a mutex lock here, as we're reading
 	 *	active_threads, and not modifying it.  We want a close
@@ -910,9 +1038,14 @@ static void thread_pool_manage(time_t now)
 	/*
 	 *	If there are too few spare threads.  Go create some more.
 	 */
-	if (spare < thread_pool.min_spare_threads) {
+	if ((thread_pool.total_threads < thread_pool.max_threads) &&
+	    (spare < thread_pool.min_spare_threads)) {
 		total = thread_pool.min_spare_threads - spare;
 
+		if ((total + thread_pool.total_threads) > thread_pool.max_threads) {
+			total = thread_pool.max_threads - thread_pool.total_threads;
+		}
+
 		DEBUG2("Threads: Spawning %d spares", total);
 
 		/*
@@ -937,21 +1070,6 @@ static void thread_pool_manage(time_t now)
 	}
 	last_cleaned = now;
 
-	/*
-	 *	Loop over the thread pool, deleting exited threads.
-	 */
-	for (handle = thread_pool.head; handle; handle = next) {
-		next = handle->next;
-
-		/*
-		 *	Maybe we've asked the thread to exit, and it
-		 *	has agreed.
-		 */
-		if (handle->status == THREAD_EXITED) {
-			delete_thread(handle);
-		}
-	}
-
 	/*
 	 *	Only delete the spare threads if sufficient time has
 	 *	passed since we last created one.  This helps to minimize
@@ -1003,27 +1121,6 @@ static void thread_pool_manage(time_t now)
 		}
 	}
 
-	/*
-	 *	If the thread has handled too many requests, then make it
-	 *	exit.
-	 */
-	if (thread_pool.max_requests_per_thread > 0) {
-		for (handle = thread_pool.head; handle; handle = next) {
-			next = handle->next;
-
-			/*
-			 *	Not handling a request, but otherwise
-			 *	live, we can kill it.
-			 */
-			if ((handle->request == NULL) &&
-			    (handle->status == THREAD_RUNNING) &&
-			    (handle->request_count > thread_pool.max_requests_per_thread)) {
-				handle->status = THREAD_CANCELLED;
-				sem_post(&thread_pool.semaphore);
-			}
-		}
-	}
-
 	/*
 	 *	Otherwise everything's kosher.  There are not too few,
 	 *	or too many spare threads.  Exit happily.
@@ -1051,6 +1148,7 @@ pid_t rad_fork(void)
 	/*
 	 *	Fork & save the PID for later reaping.
 	 */
+	pthread_mutex_lock(&thread_pool.wait_mutex);
 	child_pid = fork();
 	if (child_pid > 0) {
 		int rcode;
@@ -1061,15 +1159,20 @@ pid_t rad_fork(void)
 
 		tf->pid = child_pid;
 
-		pthread_mutex_lock(&thread_pool.wait_mutex);
 		rcode = fr_hash_table_insert(thread_pool.waiters, tf);
-		pthread_mutex_unlock(&thread_pool.wait_mutex);
 
 		if (!rcode) {
 			radlog(L_ERR, "Failed to store PID, creating what will be a zombie process %d",
 			       (int) child_pid);
+			free(tf);
 		}
 	}
+	/*
+	 *	Do not unlock in child process
+	 */
+	if(child_pid != 0 ) {
+		pthread_mutex_unlock(&thread_pool.wait_mutex);
+	}
 
 	/*
 	 *	Return whatever we were told.
