@@ -1,7 +1,7 @@
 /*
  * rlm_logintime.c
  *
- * Version:  $Id: rlm_logintime.c,v 1.17 2007/11/25 14:02:10 aland Exp $
+ * Version:  $Id: 6a26b9bad017590ddc6239f9c53d40ce49a2bfd1 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id: rlm_logintime.c,v 1.17 2007/11/25 14:02:10 aland Exp $")
+RCSID("$Id: 6a26b9bad017590ddc6239f9c53d40ce49a2bfd1 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
@@ -322,6 +322,6 @@ module_t rlm_logintime = {
 		NULL,			/* checksimul */
 		NULL,			/* pre-proxy */
 		NULL,			/* post-proxy */
-		NULL			/* post-auth */
+		logintime_authorize    	/* post-auth */
 	},
 };
